@@ -30,11 +30,16 @@ from __future__ import absolute_import
 from dbus import DBusException, Interface
 import logging
 import subprocess
-from time import sleep
 from functools import partial
 import os
 import psutil
+from six import u
 
+from autopilot.dbus_handler import (
+    get_session_bus,
+    get_system_bus,
+    get_custom_bus,
+)
 from autopilot.introspection.backends import DBusAddress
 from autopilot.introspection.constants import (
     AUTOPILOT_PATH,
@@ -50,14 +55,10 @@ from autopilot.introspection.utilities import (
     _get_bus_connections_pid,
     _pid_is_running,
 )
-from autopilot.dbus_handler import (
-    get_session_bus,
-    get_system_bus,
-    get_custom_bus,
-)
+from autopilot._timeout import Timeout
 
 
-logger = logging.getLogger(__name__)
+_logger = logging.getLogger(__name__)
 
 # Keep track of known connections during search
 connection_list = []
@@ -67,108 +68,6 @@ class ProcessSearchError(RuntimeError):
     pass
 
 
-def get_application_launcher(app_path):
-    """Return an instance of :class:`ApplicationLauncher` that knows how to
-    launch the application at 'app_path'.
-    """
-    # TODO: this is a teeny bit hacky - we call ldd to check whether this
-    # application links to certain library. We're assuming that linking to
-    # libQt* or libGtk* means the application is introspectable. This excludes
-    # any non-dynamically linked executables, which we may need to fix further
-    # down the line.
-    try:
-        ldd_output = subprocess.check_output(["ldd", app_path]).strip().lower()
-    except subprocess.CalledProcessError as e:
-        raise RuntimeError(e)
-    if 'libqtcore' in ldd_output or 'libqt5core' in ldd_output:
-        from autopilot.introspection.qt import QtApplicationLauncher
-        return QtApplicationLauncher()
-    elif 'libgtk' in ldd_output:
-        from autopilot.introspection.gtk import GtkApplicationLauncher
-        return GtkApplicationLauncher()
-    return None
-
-
-def get_application_launcher_from_string_hint(hint):
-    """Return in instance of :class:`ApplicationLauncher` given a string
-    hint."""
-    from autopilot.introspection.qt import QtApplicationLauncher
-    from autopilot.introspection.gtk import GtkApplicationLauncher
-
-    hint = hint.lower()
-    if hint == 'qt':
-        return QtApplicationLauncher()
-    elif hint == 'gtk':
-        return GtkApplicationLauncher()
-    return None
-
-
-def launch_application(launcher, application, *arguments, **kwargs):
-    """Launch an application, and return a process object.
-
-    :param launcher: An instance of the :class:`ApplicationLauncher` class to
-        prepare the environment before launching the application itself.
-    """
-
-    if not isinstance(application, basestring):
-        raise TypeError("'application' parameter must be a string.")
-    cwd = kwargs.pop('launch_dir', None)
-    capture_output = kwargs.pop('capture_output', True)
-    if kwargs:
-        raise ValueError(
-            "Unknown keyword arguments: %s." %
-            (', '.join(repr(k) for k in kwargs.keys())))
-
-    path, args = launcher.prepare_environment(application, list(arguments))
-
-    process = launch_process(
-        path,
-        args,
-        capture_output,
-        cwd=cwd
-    )
-    return process
-
-
-class ApplicationLauncher(object):
-    """A class that knows how to launch an application with a certain type of
-    introspection enabled.
-
-    """
-
-    def prepare_environment(self, app_path, arguments):
-        """Prepare the application, or environment to launch with
-        autopilot-support.
-
-        This method does nothing - it exists so child classes can override it.
-
-        The method *must* return a tuple of (*app_path*, *arguments*). Either
-        of these can be altered by this method.
-
-        """
-        raise NotImplementedError("Sub-classes must implement this method.")
-
-
-def launch_process(application, args, capture_output, **kwargs):
-    """Launch an autopilot-enabled process and return the process object."""
-    commandline = [application]
-    commandline.extend(args)
-    logger.info("Launching process: %r", commandline)
-    cap_mode = None
-    if capture_output:
-        cap_mode = subprocess.PIPE
-    process = subprocess.Popen(
-        commandline,
-        stdin=subprocess.PIPE,
-        stdout=cap_mode,
-        stderr=cap_mode,
-        close_fds=True,
-        preexec_fn=os.setsid,
-        **kwargs
-    )
-    return process
-
-
 def get_autopilot_proxy_object_for_process(
     process,
     emulator_base,
@@ -245,14 +144,7 @@ def get_proxy_object_for_existing_process(
         ``process.pid != pid``.
 
     """
-    if process is not None:
-        if pid is None:
-            pid = process.pid
-        elif pid != process.pid:
-            raise RuntimeError("Supplied PID and process.pid do not match.")
-
-    if pid is not None and not _pid_is_running(pid):
-        raise ProcessSearchError("PID %d could not be found" % pid)
+    pid = _check_process_and_pid_details(process, pid)
 
     dbus_addresses = _get_dbus_addresses_from_search_parameters(
         pid,
@@ -262,19 +154,109 @@ def get_proxy_object_for_existing_process(
         process
     )
 
-    if application_name:
-        app_name_check_fn = lambda i: get_classname_from_path(
-            i.introspection_iface.GetState('/')[0][0]) == application_name
-        dbus_addresses = filter(app_name_check_fn, dbus_addresses)
+    dbus_addresses = _maybe_filter_connections_by_app_name(
+        application_name,
+        dbus_addresses
+    )
 
     if dbus_addresses is None or len(dbus_addresses) == 0:
-        raise ProcessSearchError("Search criteria returned no results")
+        criteria_string = _get_search_criteria_string_representation(
+            pid,
+            dbus_bus,
+            connection_name,
+            process,
+            object_path,
+            application_name
+        )
+        message_string = "Search criteria (%s) returned no results" % \
+            (criteria_string)
+        raise ProcessSearchError(message_string)
     if len(dbus_addresses) > 1:
-        raise RuntimeError("Search criteria returned multiple results")
+        criteria_string = _get_search_criteria_string_representation(
+            pid,
+            dbus_bus,
+            connection_name,
+            process,
+            object_path,
+            application_name
+        )
+        message_string = "Search criteria (%s) returned multiple results" % \
+            (criteria_string)
+        raise RuntimeError(message_string)
 
     return _make_proxy_object(dbus_addresses[0], emulator_base)
 
 
+def _check_process_and_pid_details(process=None, pid=None):
+    """Do error checking on process and pid specification.
+
+    :raises RuntimeError: if both process and pid are specified, but the
+        process's 'pid' attribute is different to the pid attribute specified.
+    :raises ProcessSearchError: if the process specified is not running.
+    :returns: the pid to use in all search queries.
+
+    """
+    if process is not None:
+        if pid is None:
+            pid = process.pid
+        elif pid != process.pid:
+            raise RuntimeError("Supplied PID and process.pid do not match.")
+
+    if pid is not None and not _pid_is_running(pid):
+        raise ProcessSearchError("PID %d could not be found" % pid)
+    return pid
+
+
+def _maybe_filter_connections_by_app_name(application_name, dbus_addresses):
+    """Filter `dbus_addresses` by the application name exported, if
+    `application_name` has been specified.
+
+    :returns: a filtered list of connections.
+    """
+
+    if application_name:
+        dbus_addresses = [
+            a for a in dbus_addresses
+            if _get_application_name_from_dbus_address(a) == application_name
+        ]
+    return dbus_addresses
+
+
+def _get_application_name_from_dbus_address(dbus_address):
+    """Return the application name from a dbus_address object."""
+    return get_classname_from_path(
+        dbus_address.introspection_iface.GetState('/')[0][0]
+    )
+
+
+def _get_search_criteria_string_representation(
+        pid=None, dbus_bus=None, connection_name=None, process=None,
+        object_path=None, application_name=None):
+    """Get a string representation of the search criteria.
+
+    Used to represent the search criteria to users in error messages.
+    """
+    description_parts = []
+    if pid is not None:
+        description_parts.append(u('pid = %d') % pid)
+    if dbus_bus is not None:
+        description_parts.append(u("dbus bus = '%s'") % dbus_bus)
+    if connection_name is not None:
+        description_parts.append(
+            u("connection name = '%s'") % connection_name
+        )
+    if object_path is not None:
+        description_parts.append(u("object path = '%s'") % object_path)
+    if application_name is not None:
+        description_parts.append(
+            u("application name = '%s'") % application_name
+        )
+    if process is not None:
+        description_parts.append(u("process object = '%r'") % process)
+
+    return ", ".join(description_parts)
+
+
 def _get_dbus_addresses_from_search_parameters(
         pid, dbus_bus, connection_name, object_path, process):
     """Returns a list of :py:class: `DBusAddress` for all successfully matched
@@ -283,7 +265,7 @@ def _get_dbus_addresses_from_search_parameters(
     """
     _reset_known_connection_list()
 
-    for i in range(10):
+    for _ in Timeout.default():
         _get_child_pids.reset_cache()
         if process is not None and not _process_is_running(process):
             return_code = process.poll()
@@ -304,8 +286,6 @@ def _get_dbus_addresses_from_search_parameters(
         if len(valid_connections) >= 1:
             return [_get_dbus_address_object(name, object_path, bus) for name
                     in valid_connections]
-
-        sleep(1)
     return []
 
 
@@ -402,7 +382,7 @@ def _connection_matches_pid(bus, connection_name, pid):
             return False
         bus_pid = _get_bus_connections_pid(bus, connection_name)
     except DBusException as e:
-        logger.info(
+        _logger.info(
             "dbus.DBusException while attempting to get PID for %s: %r" %
             (connection_name, e))
         return False
@@ -466,53 +446,204 @@ _get_child_pids = _cached_get_child_pids()
 
 
 def _make_proxy_object(data_source, emulator_base):
-    """Returns a root proxy object given a DBus service name."""
-    proxy_bases = _get_proxy_object_base_classes(data_source)
-    if emulator_base is None:
-        emulator_base = type('DefaultEmulatorBase', (CustomEmulatorBase,), {})
-    proxy_bases = proxy_bases + (emulator_base, )
-    cls_name, cls_state = _get_proxy_object_class_name_and_state(data_source)
+    """Returns a root proxy object given a DBus service name.
 
-    # Merge the object hierarchy.
-    clsobj = type(str("%sBase" % cls_name), proxy_bases, {})
-
-    proxy_class = type(str(cls_name), (clsobj,), {})
+    :param data_source: The dbus backend address object we're querying.
+    :param emulator_base: The emulator base object (or None), as provided by
+        the user.
+    """
+    intro_xml = _get_introspection_xml_from_backend(data_source)
+    cls_name, path, cls_state = _get_proxy_object_class_name_and_state(
+        data_source
+    )
 
     try:
-        dbus_tuple = data_source.introspection_iface.GetState("/")[0]
-        path, state = dbus_tuple
-        return proxy_class(state, path, data_source)
-    except IndexError:
-        raise RuntimeError("Unable to find root object of %r" % proxy_class)
+        proxy_bases = _get_proxy_bases_from_introspection_xml(intro_xml)
+    except RuntimeError as e:
+        e.args = (
+            "Could not find Autopilot interface on dbus address '%s'."
+            % data_source,
+        )
+        raise e
+    proxy_bases = _extend_proxy_bases_with_emulator_base(
+        proxy_bases,
+        emulator_base
+    )
+    proxy_class = _make_proxy_class_object(cls_name, proxy_bases)
+
+    return proxy_class(cls_state, path, data_source)
 
 
-def _get_proxy_object_base_classes(backend):
-    """Return  tuple of the base classes to use when creating a proxy object
-    for the given service name & path.
+def _make_proxy_object_async(
+        data_source, emulator_base, reply_handler, error_handler):
+    """Make a proxy object for a dbus backend.
+
+    Similar to :meth:`_make_proxy_object` except this method runs
+    asynchronously and must have a reply_handler callable set. The
+    reply_handler will be called with a single argument: The proxy object.
+
+    """
+    # Note: read this function backwards!
+    #
+    # Due to the callbacks, I need to define the end of the callback chain
+    # first, so start reading from the bottom of the function upwards, and
+    # it'll make a whole lot more sense.
+
+    # Final phase: We have all the information we need, now we construct
+    # everything. This phase has no dbus calls, and so is very fast:
+    def build_proxy(introspection_xml, cls_name, path, cls_state):
+        proxy_bases = _get_proxy_bases_from_introspection_xml(
+            introspection_xml
+        )
+        proxy_bases = _extend_proxy_bases_with_emulator_base(
+            proxy_bases,
+            emulator_base
+        )
+        proxy_class = _make_proxy_class_object(cls_name, proxy_bases)
+        reply_handler(proxy_class(cls_state, path, data_source))
+
+    # Phase 2: We recieve the introspection string, and make an asynchronous
+    # dbus call to get the state information for the root of this applicaiton.
+    def get_root_state(introspection_xml):
+        _get_proxy_object_class_name_and_state(
+            data_source,
+            reply_handler=partial(build_proxy, introspection_xml),
+            error_handler=error_handler,
+        )
+
+    # Phase 1: Make an asynchronous dbus call to get the introspection xml
+    # from the data_source provided for us.
+    _get_introspection_xml_from_backend(
+        data_source,
+        reply_handler=get_root_state,
+        error_handler=error_handler
+    )
+
+
+def _get_introspection_xml_from_backend(
+        backend, reply_handler=None, error_handler=None):
+    """Get DBus Introspection xml from a backend.
+
+    :param backend: The backend object to query.
+    :param reply_handler: If set, makes a dbus async call, and the result will
+        be sent to reply_handler. This must be a callable object.
+    :param error_handler: If set, this callable will recieve any errors, and
+        the call will be made asyncronously.
+    :returns: A string containing introspection xml, if called synchronously.
+    :raises ValueError: if one, but not both of 'reply_handler' and
+        'error_handler' are set.
+
+    """
+    if callable(reply_handler) and callable(error_handler):
+        backend.dbus_introspection_iface.Introspect(
+            reply_handler=reply_handler,
+            error_handler=error_handler,
+        )
+    elif reply_handler or error_handler:
+        raise ValueError(
+            "Both 'reply_handler' and 'error_handler' must be set."
+        )
+    else:
+        return backend.dbus_introspection_iface.Introspect()
+
+
+def _get_proxy_bases_from_introspection_xml(introspection_xml):
+    """Return  tuple of the base classes to use when creating a proxy object.
+
+    Currently this works by looking for certain interface names in the XML. In
+    the future we may want to parse the XML and perform more rigerous checks.
+
+    :param introspection_xml: An xml string that describes the exported object
+        on the dbus backend. This determines which capabilities are present in
+        the backend, and therefore which base classes should be used to create
+        the proxy object.
 
     :raises: **RuntimeError** if the autopilot interface cannot be found.
 
     """
-
     bases = [ApplicationProxyObject]
+    if AP_INTROSPECTION_IFACE not in introspection_xml:
+        raise RuntimeError("Could not find Autopilot interface.")
 
-    intro_xml = backend.dbus_introspection_iface.Introspect()
-    if AP_INTROSPECTION_IFACE not in intro_xml:
-        raise RuntimeError(
-            "Could not find Autopilot interface on DBus backend '%s'" %
-            backend)
-
-    if QT_AUTOPILOT_IFACE in intro_xml:
+    if QT_AUTOPILOT_IFACE in introspection_xml:
         from autopilot.introspection.qt import QtObjectProxyMixin
         bases.append(QtObjectProxyMixin)
 
     return tuple(bases)
 
 
-def _get_proxy_object_class_name_and_state(backend):
-    """Return the class name and root state dictionary."""
-    object_path, object_state = backend.introspection_iface.GetState("/")[0]
-    return get_classname_from_path(object_path), object_state
+def _extend_proxy_bases_with_emulator_base(proxy_bases, emulator_base):
+    if emulator_base is None:
+        emulator_base = type('DefaultEmulatorBase', (CustomEmulatorBase,), {})
+    return proxy_bases + (emulator_base, )
+
+
+def _get_proxy_object_class_name_and_state(
+        backend, reply_handler=None, error_handler=None):
+    """Get details about this autopilot backend via a dbus GetState call.
+
+    :param reply_handler: A callable that must accept three positional
+        arguments, which correspond to the return value of this function when
+        called synchronously.
+
+    :param error_handler: A callable which will recieve any dbus errors, should
+        they occur.
+
+    :raises ValueError: if one, but not both of reply_handler and error_handler
+        are set.
+
+    :returns: A tuple containing the class name of the root of the
+        introspection tree, the full path to the root of the introspection
+        tree, and the state dictionary of the root node in the introspection
+        tree.
+
+    """
+    if callable(reply_handler) and callable(error_handler):
+        # Async call:
+        # Since we get an array of state, and we only care about the first one
+        # we use a lambda to unpack it and get the details we want.
+        backend.introspection_iface.GetState(
+            "/",
+            reply_handler=lambda r: reply_handler(
+                *_get_details_from_state_data(r[0])
+            ),
+            error_handler=error_handler,
+        )
+    elif reply_handler or error_handler:
+        raise ValueError(
+            "Both 'reply_handler' and 'error_handler' must be set."
+        )
+    else:
+        # Sync call
+        state = backend.introspection_iface.GetState("/")[0]
+        return _get_details_from_state_data(state)
+
+
+def _get_details_from_state_data(state_data):
+    """Get details from a state data array.
+
+    Returns class name, path, and state dictionary.
+    """
+    object_path, object_state = state_data
+    return (
+        get_classname_from_path(object_path),
+        object_path,
+        object_state,
+    )
+
+
+def _make_proxy_class_object(cls_name, proxy_bases):
+    """Return a class object for a proxy.
+
+    :param cls_name: The name of the class to be created. This is usually the
+        type as returned over the wire from an autopilot backend.
+    :param proxy_bases: A list of base classes this proxy class should derive
+        from. This determines the specific abilities of the new proxy.
+
+    """
+    clsobj = type(str("%sBase" % cls_name), proxy_bases, {})
+    proxy_class = type(str(cls_name), (clsobj,), {})
+    return proxy_class
 
 
 class ApplicationProxyObject(DBusIntrospectionObject):
