@@ -28,17 +28,23 @@ module is the DBusIntrospectableObject class.
 from __future__ import absolute_import
 
 from contextlib import contextmanager
+import sys
 import logging
 import re
-from testtools.matchers import Equals
-from time import sleep
+import six
 from uuid import uuid4
 
-from autopilot.utilities import Timer, get_debug_logger
+from autopilot.introspection.types import create_value_instance
+from autopilot.introspection.utilities import translate_state_keys
+from autopilot.utilities import (
+    get_debug_logger,
+    sleep,
+    Timer,
+)
 
 
 _object_registry = {}
-logger = logging.getLogger(__name__)
+_logger = logging.getLogger(__name__)
 
 
 class StateNotFoundError(RuntimeError):
@@ -49,11 +55,13 @@ class StateNotFoundError(RuntimeError):
     not yet created) the object you are trying to access in autopilot. This
     typically happens for a number of possible reasons:
 
-     * The UI widget you are trying to access with
-        :py:met:`DBusIntrospectionObject.select_single` or
-        :py:met:`DBusIntrospectionObject.select_single` does not exist yet.
+    * The UI widget you are trying to access with
+      :py:meth:`~DBusIntrospectionObject.select_single` or
+      :py:meth:`~DBusIntrospectionObject.wait_select_single` or
+      :py:meth:`~DBusIntrospectionObject.select_many` does not exist yet.
+
     * The UI widget you are trying to access has been destroyed by the
-        application.
+      application.
 
     """
 
@@ -69,22 +77,25 @@ class StateNotFoundError(RuntimeError):
 
         if class_name is None:
             self._message = \
-                u"State not found with filters {}.".format(
+                u"Object not found with properties {}.".format(
                     repr(filters)
                 )
         elif not filters:
-            self._message = u"State not found for class '{}'.".format(
+            self._message = u"Object not found with name '{}'.".format(
                 class_name
             )
         else:
             self._message = \
-                u"State not found for class '{}' and filters {}.".format(
+                u"Object not found with name '{}' and properties {}.".format(
                     class_name,
                     repr(filters)
                 )
 
     def __str__(self):
-        return self._message.encode('utf8')
+        if six.PY3:
+            return self._message
+        else:
+            return self._message.encode('utf8')
 
     def __unicode__(self):
         return self._message
@@ -111,21 +122,15 @@ class IntrospectableObjectMetaclass(type):
         return class_object
 
 
-def translate_state_keys(state_dict):
-    """Translates the *state_dict* passed in so the keys are usable as python
-    attributes."""
-    return {k.replace('-', '_'): v for k, v in state_dict.iteritems()}
-
-
 def get_classname_from_path(object_path):
     return object_path.split("/")[-1]
 
 
-def object_passes_filters(instance, **kwargs):
+def _object_passes_filters(instance, **kwargs):
     """Return true if *instance* satisifies all the filters present in
     kwargs."""
     with instance.no_automatic_refreshing():
-        for attr, val in kwargs.iteritems():
+        for attr, val in kwargs.items():
             if not hasattr(instance, attr) or getattr(instance, attr) != val:
                 # Either attribute is not present, or is present but with
                 # the wrong value - don't add this instance to the results
@@ -134,7 +139,14 @@ def object_passes_filters(instance, **kwargs):
     return True
 
 
-class DBusIntrospectionObject(object):
+DBusIntrospectionObjectBase = IntrospectableObjectMetaclass(
+    'DBusIntrospectionObjectBase',
+    (object,),
+    {}
+)
+
+
+class DBusIntrospectionObject(DBusIntrospectionObjectBase):
     """A class that supports transparent data retrieval from the application
     under test.
 
@@ -145,13 +157,12 @@ class DBusIntrospectionObject(object):
 
     """
 
-    __metaclass__ = IntrospectableObjectMetaclass
-
     def __init__(self, state_dict, path, backend):
         self.__state = {}
         self.__refresh_on_attribute = True
         self._set_properties(state_dict)
         self._path = path
+        self._poll_time = 10
         self._backend = backend
 
     def _set_properties(self, state_dict):
@@ -163,105 +174,34 @@ class DBusIntrospectionObject(object):
 
         """
         self.__state = {}
-        for key, value in translate_state_keys(state_dict).iteritems():
+        for key, value in translate_state_keys(state_dict).items():
             # don't store id in state dictionary -make it a proper instance
             # attribute
             if key == 'id':
-                self.id = value
-            self.__state[key] = self._make_attribute(key, value)
-
-    def _make_attribute(self, name, value):
-        """Make an attribute for *value*, patched with the wait_for
-        function."""
-
-        def wait_for(self, expected_value, timeout=10):
-            """Wait up to 10 seconds for our value to change to
-            *expected_value*.
-
-            *expected_value* can be a testtools.matcher. Matcher subclass (like
-            LessThan, for example), or an ordinary value.
-
-            This works by refreshing the value using repeated dbus calls.
-
-            :raises: **RuntimeError** if the attribute was not equal to the
-             expected value after 10 seconds.
-
-            """
-            # It's guaranteed that our value is up to date, since __getattr__
-            # calls refresh_state. This if statement stops us waiting if the
-            # value is already what we expect:
-            if self == expected_value:
-                return
-
-            def make_unicode(value):
-                if isinstance(value, str):
-                    return unicode(value.decode('utf8'))
-                return value
-
-            if hasattr(expected_value, 'expected'):
-                expected_value.expected = make_unicode(expected_value.expected)
-
-            # unfortunately not all testtools matchers derive from the Matcher
-            # class, so we can't use issubclass, isinstance for this:
-            match_fun = getattr(expected_value, 'match', None)
-            is_matcher = match_fun and callable(match_fun)
-            if not is_matcher:
-                expected_value = Equals(expected_value)
-
-            time_left = timeout
-            while True:
-                _, new_state = self.parent.get_new_state()
-                new_state = translate_state_keys(new_state)
-                new_value = make_unicode(new_state[self.name])
-                # Support for testtools.matcher classes:
-                mismatch = expected_value.match(new_value)
-                if mismatch:
-                    failure_msg = mismatch.describe()
-                else:
-                    self.parent._set_properties(new_state)
-                    return
-
-                if time_left >= 1:
-                    sleep(1)
-                    time_left -= 1
-                else:
-                    sleep(time_left)
-                    break
-
-            raise AssertionError(
-                "After %.1f seconds test on %s.%s failed: %s" % (
-                    timeout, self.parent.__class__.__name__, self.name,
-                    failure_msg))
-
-        # This looks like magic, but it's really not. We're creating a new type
-        # on the fly that derives from the type of 'value' with a couple of
-        # extra attributes: wait_for is the wait_for method above. 'parent' and
-        # 'name' are needed by the wait_for method.
-        #
-        # We can't use traditional meta-classes here, since the type we're
-        # deriving from is only known at call time, not at parse time (we could
-        # override __call__ in the meta class, but that doesn't buy us anything
-        # extra).
-        #
-        # A better way to do this would be with functools.partial, which I
-        # tried initially, but doesn't work well with bound methods.
-        t = type(value)
-        attrs = {'wait_for': wait_for, 'parent': self, 'name': name}
-        return type(t.__name__, (t,), attrs)(value)
+                self.id = int(value[1])
+            try:
+                self.__state[key] = create_value_instance(value, self, key)
+            except ValueError as e:
+                _logger.warning(
+                    "While constructing attribute '%s.%s': %s",
+                    self.__class__.__name__,
+                    key,
+                    str(e)
+                )
 
     def get_children_by_type(self, desired_type, **kwargs):
         """Get a list of children of the specified type.
 
         Keyword arguments can be used to restrict returned instances. For
-        example:
+        example::
 
-        >>> get_children_by_type(Launcher, monitor=1)
+            get_children_by_type('Launcher', monitor=1)
 
         will return only Launcher instances that have an attribute 'monitor'
         that is equal to 1. The type can also be specified as a string, which
-        is useful if there is no emulator class specified:
+        is useful if there is no emulator class specified::
 
-        >>> get_children_by_type('Launcher', monitor=1)
+            get_children_by_type('Launcher', monitor=1)
 
         Note however that if you pass a string, and there is an emulator class
         defined, autopilot will not use it.
@@ -271,7 +211,7 @@ class DBusIntrospectionObject(object):
             custom emulators)
 
         .. seealso::
-            Tutorial Section :ref:`custom_emulators`
+            Tutorial Section :ref:`custom_proxy_classes`
 
         """
         #TODO: if kwargs has exactly one item in it we should specify the
@@ -282,14 +222,14 @@ class DBusIntrospectionObject(object):
         result = []
         for instance in instances:
             # Skip items that are not instances of the desired type:
-            if isinstance(desired_type, basestring):
+            if isinstance(desired_type, six.string_types):
                 if instance.__class__.__name__ != desired_type:
                     continue
             elif not isinstance(instance, desired_type):
                 continue
 
             #skip instances that fail attribute check:
-            if object_passes_filters(instance, **kwargs):
+            if _object_passes_filters(instance, **kwargs):
                 result.append(instance)
         return result
 
@@ -319,13 +259,27 @@ class DBusIntrospectionObject(object):
         :meth:`select_many`.
 
         """
-        self.refresh_state()
-
+        # Thomi: 2014-03-20: There used to be a call to 'self.refresh_state()'
+        # here. That's not needed, since the only thing we use is the proxy
+        # path, which isn't affected by the current state.
         query = self.get_class_query_string() + "/*"
         state_dicts = self.get_state_by_path(query)
         children = [self.make_introspection_object(i) for i in state_dicts]
         return children
 
+    def get_parent(self):
+        """Returns the parent of this object.
+
+        If this object has no parent (i.e.- it is the root of the introspection
+        tree). Then it returns itself.
+
+        """
+        query = self.get_class_query_string() + "/.."
+        parent_state_dicts = self.get_state_by_path(query)
+
+        parent = self.make_introspection_object(parent_state_dicts[0])
+        return parent
+
     def select_single(self, type_name='*', **kwargs):
         """Get a single node from the introspection tree, with type equal to
         *type_name* and (optionally) matching the keyword filters present in
@@ -344,7 +298,8 @@ class DBusIntrospectionObject(object):
             app.select_single('QPushButton', objectName='clickme')
             # returns a QPushButton whose 'objectName' property is 'clickme'.
 
-        If nothing is returned from the query, this method returns None.
+        If nothing is returned from the query, this method raises
+        StateNotFoundError.
 
         :param type_name: Either a string naming the type you want, or a class
             of the appropriate type (the latter case is for overridden emulator
@@ -356,17 +311,73 @@ class DBusIntrospectionObject(object):
         :raises TypeError: if neither *type_name* or keyword filters are
             provided.
 
+        :raises StateNotFoundError: if the requested object was not found.
+
         .. seealso::
-            Tutorial Section :ref:`custom_emulators`
+            Tutorial Section :ref:`custom_proxy_classes`
 
         """
         instances = self.select_many(type_name, **kwargs)
         if len(instances) > 1:
             raise ValueError("More than one item was returned for query")
         if not instances:
-            return None
+            raise StateNotFoundError(type_name, **kwargs)
         return instances[0]
 
+    def wait_select_single(self, type_name='*', **kwargs):
+        """Get a proxy object matching some search criteria, retrying if no
+        object is found until a timeout is reached.
+
+        This method is identical to the :meth:`select_single` method, except
+        that this method will poll the application under test for 10 seconds
+        in the event that the search criteria does not match anything.
+
+        This method will return single proxy object from the introspection
+        tree, with type equal to *type_name* and (optionally) matching the
+        keyword filters present in *kwargs*.
+
+        You must specify either *type_name*, keyword filters or both.
+
+        This method searches recursively from the proxy object this method is
+        called on. Calling :meth:`select_single` on the application (root)
+        proxy object will search the entire tree. Calling
+        :meth:`select_single` on an object in the tree will only search it's
+        descendants.
+
+        Example usage::
+
+            app.wait_select_single('QPushButton', objectName='clickme')
+            # returns a QPushButton whose 'objectName' property is 'clickme'.
+            # will poll the application until such an object exists, or will
+            # raise StateNotFoundError after 10 seconds.
+
+        If nothing is returned from the query, this method raises
+        StateNotFoundError after 10 seconds.
+
+        :param type_name: Either a string naming the type you want, or a class
+            of the appropriate type (the latter case is for overridden emulator
+            classes).
+
+        :raises ValueError: if the query returns more than one item. *If
+            you want more than one item, use select_many instead*.
+
+        :raises TypeError: if neither *type_name* or keyword filters are
+            provided.
+
+        :raises StateNotFoundError: if the requested object was not found.
+
+        .. seealso::
+            Tutorial Section :ref:`custom_proxy_classes`
+
+        """
+        for i in range(self._poll_time):
+            try:
+                return self.select_single(type_name, **kwargs)
+            except StateNotFoundError:
+                if i == self._poll_time - 1:
+                    raise
+                sleep(1)
+
     def select_many(self, type_name='*', **kwargs):
         """Get a list of nodes from the introspection tree, with type equal to
         *type_name* and (optionally) matching the keyword filters present in
@@ -384,11 +395,18 @@ class DBusIntrospectionObject(object):
             app.select_many('QPushButton', enabled=True)
             # returns a list of QPushButtons that are enabled.
 
-        As mentioned above, this method searches the object tree recurseivly::
+        As mentioned above, this method searches the object tree recursively::
+
             file_menu = app.select_one('QMenu', title='File')
             file_menu.select_many('QAction')
             # returns a list of QAction objects who appear below file_menu in
-            the object tree.
+            # the object tree.
+
+        .. warning::
+            The order in which objects are returned is not guaranteed. It is
+            bad practise to write tests that depend on the order in which
+            this method returns objects. (see :ref:`object_ordering` for more
+            information).
 
         If you only want to get one item, use :meth:`select_single` instead.
 
@@ -396,11 +414,11 @@ class DBusIntrospectionObject(object):
             of the appropriate type (the latter case is for overridden emulator
             classes).
 
-        :raises: **TypeError** if neither *type_name* or keyword filters are
+        :raises TypeError: if neither *type_name* or keyword filters are
             provided.
 
         .. seealso::
-            Tutorial Section :ref:`custom_emulators`
+            Tutorial Section :ref:`custom_proxy_classes`
 
         """
         if not isinstance(type_name, str) and issubclass(
@@ -410,12 +428,13 @@ class DBusIntrospectionObject(object):
         if type_name == "*" and not kwargs:
             raise TypeError("You must specify either a type name or a filter.")
 
-        logger.debug(
+        _logger.debug(
             "Selecting objects of %s with attributes: %r",
             'any type' if type_name == '*' else 'type ' + type_name, kwargs)
 
-        first_param = ''
-        for k, v in kwargs.iteritems():
+        server_side_filters = []
+        client_side_filters = {}
+        for k, v in kwargs.items():
             # LP Bug 1209029: The XPathSelect protocol does not allow all valid
             # node names or values. We need to decide here whether the filter
             # parameters are going to work on the backend or not. If not, we
@@ -423,25 +442,33 @@ class DBusIntrospectionObject(object):
             # _is_valid_server_side_filter_param function (below) for the
             # specific requirements.
             if _is_valid_server_side_filter_param(k, v):
-                first_param = '[{}={}]'.format(k, v)
-                kwargs.pop(k)
-                break
-        query_path = "%s//%s%s" % (self.get_class_query_string(),
-                                   type_name,
-                                   first_param)
+                server_side_filters.append(
+                    _get_filter_string_for_key_value_pair(k, v)
+                )
+            else:
+                client_side_filters[k] = v
+        filter_str = '[{}]'.format(','.join(server_side_filters)) \
+            if server_side_filters else ""
+        query_path = "%s//%s%s" % (
+            self.get_class_query_string(),
+            type_name,
+            filter_str
+        )
 
         state_dicts = self.get_state_by_path(query_path)
         instances = [self.make_introspection_object(i) for i in state_dicts]
-        return filter(lambda i: object_passes_filters(i, **kwargs), instances)
+        return [i for i in instances
+                if _object_passes_filters(i, **client_side_filters)]
 
     def refresh_state(self):
-        """Refreshes the object's state from unity.
+        """Refreshes the object's state.
 
         You should probably never have to call this directly. Autopilot
         automatically retrieves new state every time this object's attributes
         are read.
 
-        :raises: **StateNotFound** if the object in unity has been destroyed.
+        :raises StateNotFound: if the object in the application under test
+            has been destroyed.
 
         """
         _, new_state = self.get_new_state()
@@ -451,9 +478,9 @@ class DBusIntrospectionObject(object):
         """Get all instances of this class that exist within the Application
         state tree.
 
-        For example, to get all the LauncherIcon instances:
+        For example, to get all the LauncherIcon instances::
 
-        >>> icons = LauncherIcon.get_all_instances()
+            icons = LauncherIcon.get_all_instances()
 
         .. warning::
             Using this method is slow - it requires a complete scan of the
@@ -478,7 +505,7 @@ class DBusIntrospectionObject(object):
         """
         instances = self.get_state_by_path("/")
         if len(instances) != 1:
-            logger.error("Could not retrieve root object.")
+            _logger.error("Could not retrieve root object.")
             return None
         return self.make_introspection_object(instances[0])
 
@@ -504,17 +531,17 @@ class DBusIntrospectionObject(object):
 
         :param piece: an XPath-like query that specifies which bit of the tree
             you want to look at.
-        :raises: **TypeError** on invalid *piece* parameter.
+        :raises TypeError: on invalid *piece* parameter.
 
         """
-        if not isinstance(piece, basestring):
+        if not isinstance(piece, six.string_types):
             raise TypeError(
                 "XPath query must be a string, not %r", type(piece))
 
         with Timer("GetState %s" % piece):
             data = self._backend.introspection_iface.GetState(piece)
             if len(data) > 15:
-                logger.warning(
+                _logger.warning(
                     "Your query '%s' returned a lot of data (%d items). This "
                     "is likely to be slow. You may want to consider optimising"
                     " your query to return fewer items.",
@@ -536,6 +563,29 @@ class DBusIntrospectionObject(object):
         except IndexError:
             raise StateNotFoundError(self.__class__.__name__, id=self.id)
 
+    def wait_until_destroyed(self, timeout=10):
+        """Block until this object is destroyed in the application.
+
+        Block until the object this instance is a proxy for has been destroyed
+        in the applicaiton under test. This is commonly used to wait until a
+        UI component has been destroyed.
+
+        :param timeout: The number of seconds to wait for the object to be
+            destroyed. If not specified, defaults to 10 seconds.
+        :raises RuntimeError: if the method timed out.
+
+        """
+        for i in range(timeout):
+            try:
+                self.get_new_state()
+                sleep(1)
+            except StateNotFoundError:
+                return
+        else:
+            raise RuntimeError(
+                "Object was not destroyed after %d seconds" % timeout
+            )
+
     def get_class_query_string(self):
         """Get the XPath query string required to refresh this class's
         state."""
@@ -549,25 +599,68 @@ class DBusIntrospectionObject(object):
         (path, state_dict).
 
         This only works for classes that derive from DBusIntrospectionObject.
+
+        :returns: A proxy object that derives from DBusIntrospectionObject
+        :raises ValueError: if more than one class is appropriate for this
+                 dbus_tuple
+
         """
         path, state = dbus_tuple
-        name = get_classname_from_path(path)
+        class_object = _get_proxy_object_class(
+            _object_registry[self._id], type(self), path, state)
+        return class_object(state, path, self._backend)
+
+    def print_tree(self, output=None, maxdepth=None, _curdepth=0):
+        """Print properties of the object and its children to a stream.
+
+        When writing new tests, this can be called when it is too difficult to
+        find the widget or property that you are interested in in "vis".
+
+        .. warning:: Do not use this in production tests, this is expensive and
+            not at all appropriate for actual testing. Only call this
+            temporarily and replace with proper select_single/select_many
+            calls.
+
+        :param output: A file object or path name where the output will be
+            written to. If not given, write to stdout.
+
+        :param maxdepth: If given, limit the maximum recursion level to that
+            number, i. e. only print children which have at most maxdepth-1
+            intermediate parents.
+
+        """
+        if maxdepth is not None and _curdepth > maxdepth:
+            return
+
+        indent = "  " * _curdepth
+        if output is None:
+            output = sys.stdout
+        elif isinstance(output, six.string_types):
+            output = open(output, 'w')
+
+        # print path
+        if _curdepth > 0:
+            output.write("\n")
+        output.write("%s== %s ==\n" % (indent, self._path))
+        # Thomi 2014-03-20: For all levels other than the top level, we can
+        # avoid an entire dbus round trip if we grab the underlying property
+        # dictionary directly. We can do this since the print_tree function
+        # that called us will have retrieved us via a call to get_children(),
+        # which gets the latest state anyway.
+        if _curdepth > 0:
+            properties = self.__state.copy()
+        else:
+            properties = self.get_properties()
+        # print properties
         try:
-            class_type = _object_registry[self._id][name]
-        except KeyError:
-            get_debug_logger().warning(
-                "Generating introspection instance for type '%s' based on "
-                "generic class.", name)
-            # we want the object to inherit from the custom emulator base, not
-            # the object class that is doing the selecting
-            for base in type(self).__bases__:
-                if issubclass(base, CustomEmulatorBase):
-                    base_class = base
-                    break
-            else:
-                base_class = type(self)
-            class_type = type(str(name), (base_class,), {})
-        return class_type(state, path, self._backend)
+            for key in sorted(properties.keys()):
+                output.write("%s%s: %r\n" % (indent, key, properties[key]))
+            # print children
+            if maxdepth is None or _curdepth < maxdepth:
+                for c in self.get_children():
+                    c.print_tree(output, maxdepth, _curdepth + 1)
+        except StateNotFoundError as error:
+            output.write("%sError: %s\n" % (indent, error))
 
     @contextmanager
     def no_automatic_refreshing(self):
@@ -576,8 +669,8 @@ class DBusIntrospectionObject(object):
 
         Example usage:
 
-        >>> with instance.no_automatic_refreshing():
-            # access lots of attributes.
+            with instance.no_automatic_refreshing():
+                # access lots of attributes.
 
         This can be useful if you need to check lots of attributes in a tight
         loop, or if you want to atomicaly check several attributes at once.
@@ -589,16 +682,151 @@ class DBusIntrospectionObject(object):
         finally:
             self.__refresh_on_attribute = True
 
+    @classmethod
+    def validate_dbus_object(cls, path, _state):
+        """Return whether this class is the appropriate proxy object class for
+        a given dbus path and state.
+
+        The default version matches the name of the dbus object and the class.
+        Subclasses of CustomProxyObject can override it to define a different
+        validation method.
+
+        :param path: The dbus path of the object to check
+        :param state: The dbus state dict of the object to check
+                      (ignored in default implementation)
+        :returns: Whether this class is appropriate for the dbus object
+
+        """
+        name = get_classname_from_path(path)
+        return cls.__name__ == name
+
 
 def _is_valid_server_side_filter_param(key, value):
     """Return True if the key and value parameters are valid for server-side
     processing.
 
     """
-    return (
-        isinstance(value, str) and
-        re.match(r'^[a-zA-Z0-9_\-]+( [a-zA-Z0-9_\-])*$', key) is not None and
-        re.match(r'^[a-zA-Z0-9_\-]+( [a-zA-Z0-9_\-])*$', value) is not None)
+    key_is_valid = re.match(
+        r'^[a-zA-Z0-9_\-]+( [a-zA-Z0-9_\-])*$',
+        key
+    ) is not None
+
+    if type(value) == int:
+        return key_is_valid and (-2**31 <= value <= 2**31 - 1)
+
+    elif type(value) == bool:
+        return key_is_valid
+
+    elif type(value) == six.binary_type:
+        return key_is_valid
+
+    elif type(value) == six.text_type:
+        try:
+            value.encode('ascii')
+            return key_is_valid
+        except UnicodeEncodeError:
+            pass
+    return False
+
+
+def _get_filter_string_for_key_value_pair(key, value):
+    """Return a string representing the filter query for this key/value pair.
+
+    The value must be suitable for server-side filtering. Raises ValueError if
+    this is not the case.
+
+    """
+    if isinstance(value, six.text_type):
+        if six.PY3:
+            escaped_value = value.encode("unicode_escape")\
+                .decode('ASCII')\
+                .replace("'", "\\'")
+        else:
+            escaped_value = value.encode('utf-8').encode("string_escape")
+            # note: string_escape codec escapes "'" but not '"'...
+            escaped_value = escaped_value.replace('"', r'\"')
+        return '{}="{}"'.format(key, escaped_value)
+    elif isinstance(value, six.binary_type):
+        if six.PY3:
+            escaped_value = value.decode('utf-8')\
+                .encode("unicode_escape")\
+                .decode('ASCII')\
+                .replace("'", "\\'")
+        else:
+            escaped_value = value.encode("string_escape")
+            # note: string_escape codec escapes "'" but not '"'...
+            escaped_value = escaped_value.replace('"', r'\"')
+        return '{}="{}"'.format(key, escaped_value)
+    elif isinstance(value, int) or isinstance(value, bool):
+        return "{}={}".format(key, repr(value))
+    else:
+        raise ValueError("Unsupported value type: {}".format(type(value)))
+
+
+def _get_proxy_object_class(proxy_class_dict, default_class, path, state):
+    """Return a custom proxy class, either from the list or the default.
+
+    Use helper functions to check the class list or return the default.
+    :param proxy_class_dict: dict of proxy classes to try
+    :param default_class: default class to use if nothing in dict matches
+    :param path: dbus path
+    :param state: dbus state
+    :returns: appropriate custom proxy class
+    :raises ValueError: if more than one class in the dict matches
+
+    """
+    class_type = _try_custom_proxy_classes(proxy_class_dict, path, state)
+    if class_type:
+        return class_type
+    return _get_default_proxy_class(default_class,
+                                    get_classname_from_path(path))
+
+
+def _try_custom_proxy_classes(proxy_class_dict, path, state):
+    """Identify which custom proxy class matches the dbus path and state.
+
+    If more than one class in proxy_class_dict matches, raise an exception.
+    :param proxy_class_dict: dict of proxy classes to try
+    :param path: dbus path
+    :param state: dbus state dict
+    :returns: matching custom proxy class
+    :raises ValueError: if more than one class matches
+
+    """
+    possible_classes = [c for c in proxy_class_dict.values() if
+                        c.validate_dbus_object(path, state)]
+    if len(possible_classes) > 1:
+        raise ValueError(
+            'More than one custom proxy class matches this object: '
+            'Matching classes are: %s. State is %s.  Path is %s.'
+            ','.join([repr(c) for c in possible_classes]),
+            repr(state),
+            path)
+    if len(possible_classes) == 1:
+        return possible_classes[0]
+    return None
+
+
+def _get_default_proxy_class(default_class, name):
+    """Return a custom proxy object class of the default or a base class.
+
+    We want the object to inherit from CustomEmulatorBase, not the object
+    class that is doing the selecting.
+    :param default_class: default class to use if no bases match
+    :param name: name of new class
+    :returns: custom proxy object class
+
+    """
+    get_debug_logger().warning(
+        "Generating introspection instance for type '%s' based on generic "
+        "class.", name)
+    for base in default_class.__bases__:
+        if issubclass(base, CustomEmulatorBase):
+            base_class = base
+            break
+    else:
+        base_class = default_class
+    return type(str(name), (base_class,), {})
 
 
 class _CustomEmulatorMeta(IntrospectableObjectMetaclass):
@@ -616,15 +844,14 @@ class _CustomEmulatorMeta(IntrospectableObjectMetaclass):
                 d['_id'] = uuid4()
         return super(_CustomEmulatorMeta, cls).__new__(cls, name, bases, d)
 
-
-class CustomEmulatorBase(DBusIntrospectionObject):
-
+CustomEmulatorBase = _CustomEmulatorMeta('CustomEmulatorBase',
+                                         (DBusIntrospectionObject, ),
+                                         {})
+CustomEmulatorBase.__doc__ = \
     """This class must be used as a base class for any custom emulators defined
     within a test case.
 
     .. seealso::
-        Tutorial Section :ref:`custom_emulators`
+        Tutorial Section :ref:`custom_proxy_classes`
             Information on how to write custom emulators.
     """
-
-    __metaclass__ = _CustomEmulatorMeta
