@@ -1,7 +1,7 @@
 # -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
 #
 # Autopilot Functional Test Tool
-# Copyright (C) 2012-2013 Canonical
+# Copyright (C) 2012-2014 Canonical
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -28,13 +28,15 @@ from gi.repository import Gio
 from gi.repository import GLib
 import logging
 import os
-from time import sleep
 from Xlib import display, X, protocol
 from subprocess import check_output, CalledProcessError, call
 
 from autopilot.dbus_handler import get_session_bus
-from autopilot.utilities import addCleanup, Silence
-
+from autopilot.utilities import (
+    addCleanup,
+    Silence,
+)
+from autopilot._timeout import Timeout
 from autopilot.process import (
     ProcessManager as ProcessManagerBase,
     Application as ApplicationBase,
@@ -45,7 +47,7 @@ from autopilot.process import (
 _BAMF_BUS_NAME = 'org.ayatana.bamf'
 _X_DISPLAY = None
 
-logger = logging.getLogger(__name__)
+_logger = logging.getLogger(__name__)
 
 
 def get_display():
@@ -150,11 +152,11 @@ class ProcessManager(ProcessManagerBase):
         if locale:
             os.putenv("LC_ALL", locale)
             addCleanup(os.unsetenv, "LC_ALL")
-            logger.info(
+            _logger.info(
                 "Starting application '%s' with files %r in locale %s",
                 app_name, files, locale)
         else:
-            logger.info(
+            _logger.info(
                 "Starting application '%s' with files %r", app_name, files)
 
         app = self.KNOWN_APPS[app_name]
@@ -162,19 +164,18 @@ class ProcessManager(ProcessManagerBase):
         apps = self.get_running_applications_by_desktop_file(
             app['desktop-file'])
 
-        for i in range(10):
+        for _ in Timeout.default():
             try:
                 new_windows = []
                 [new_windows.extend(a.get_windows()) for a in apps]
                 filter_fn = lambda w: w.x_id not in [
                     c.x_id for c in existing_windows]
-                new_wins = filter(filter_fn, new_windows)
+                new_wins = list(filter(filter_fn, new_windows))
                 if new_wins:
                     assert len(new_wins) == 1
                     return new_wins[0]
-            except DBusException:
+            except dbus.DBusException:
                 pass
-            sleep(1)
         return None
 
     def get_open_windows_by_application(self, app_name):
@@ -199,7 +200,7 @@ class ProcessManager(ProcessManagerBase):
             if len(pids):
                 call(["kill"] + pids)
         except CalledProcessError:
-            logger.warning(
+            _logger.warning(
                 "Tried to close applicaton '%s' but it wasn't running.",
                 app_name)
 
@@ -223,7 +224,7 @@ class ProcessManager(ProcessManagerBase):
         apps = [Application(p) for p in
                 self.matcher_interface.RunningApplications()]
         if user_visible_only:
-            return filter(_filter_user_visible, apps)
+            return list(filter(_filter_user_visible, apps))
         return apps
 
     def get_running_applications_by_desktop_file(self, desktop_file):
@@ -255,7 +256,7 @@ class ProcessManager(ProcessManagerBase):
         windows = [Window(w) for w in
                    self.matcher_interface.WindowStackForMonitor(-1)]
         if user_visible_only:
-            windows = filter(_filter_user_visible, windows)
+            windows = list(filter(_filter_user_visible, windows))
         # Now sort on stacking order.
         # We explicitly convert to a list from an iterator since tests
         # frequently try and use len() on return values from these methods.
@@ -323,14 +324,29 @@ class ProcessManager(ProcessManagerBase):
         """
         if type(files) is not list:
             raise TypeError("files must be a list.")
-        proc = Gio.DesktopAppInfo.new(desktop_file)
-        # FIXME: second item is a GEerror
-        proc.launch_uris(files, None)
+        proc = _launch_application(desktop_file, files)
         if wait:
             self.wait_until_application_is_running(desktop_file, 10)
         return proc
 
 
+def _launch_application(desktop_file, files):
+    proc = Gio.DesktopAppInfo.new(desktop_file)
+    # simple launch_uris() uses GLib.SpawnFlags.SEARCH_PATH by default only,
+    # but this inherits stdout; we don't want that as it hangs when tee'ing
+    # autopilot output into a file.
+    # Instead of depending on a newer version of gir/glib attempt to use the
+    # newer verison (i.e. launch_uris_as_manager works) and fall back on using
+    # the simple launch_uris
+    try:
+        proc.launch_uris_as_manager(
+            files, None,
+            GLib.SpawnFlags.SEARCH_PATH | GLib.SpawnFlags.STDOUT_TO_DEV_NULL,
+            None, None, None, None)
+    except TypeError:
+        proc.launch_uris(files, None)
+
+
 class Application(ApplicationBase):
     """Represents an application, with information as returned by Bamf.
 
@@ -349,8 +365,8 @@ class Application(ApplicationBase):
                 self._app_proxy, 'org.ayatana.bamf.view')
             self._app_iface = dbus.Interface(
                 self._app_proxy, 'org.ayatana.bamf.application')
-        except dbus.DBusException, e:
-            e.message += 'bamf_app_path=%r' % (bamf_app_path)
+        except dbus.DBusException as e:
+            e.args += ('bamf_app_path=%r' % (bamf_app_path),)
             raise
 
     @property
@@ -413,6 +429,8 @@ class Application(ApplicationBase):
         return "<Application '%s'>" % (self.name)
 
     def __eq__(self, other):
+        if other is None:
+            return False
         return self.desktop_file == other.desktop_file
 
 
@@ -482,12 +500,12 @@ class Window(WindowBase):
         # builder.
         from gi import require_version
         require_version('GdkX11', '3.0')
-        from gi.repository import GdkX11
+        from gi.repository import Gdk, GdkX11
         # FIXME: We need to use the gdk window here to get the real coordinates
         geometry = self._x_win.get_geometry()
         origin = GdkX11.X11Window.foreign_new_for_display(
-            get_display(), self._xid).get_origin()
-        return (origin[0], origin[1], geometry.width, geometry.height)
+            Gdk.Display().get_default(), self._xid).get_origin()
+        return (origin[1], origin[2], geometry.width, geometry.height)
 
     @property
     def is_maximized(self):
@@ -611,5 +629,18 @@ class Window(WindowBase):
         """Return a list of strings representing the current window state."""
 
         get_display().sync()
-        return map(get_display().get_atom_name, self._getProperty(
-            '_NET_WM_STATE'))
+        return [get_display().get_atom_name(p)
+                for p in self._getProperty('_NET_WM_STATE')]
+
+    def resize(self, width, height):
+        """Resize the window.
+
+        :param width: The new width for the window.
+        :param height: The new height for the window.
+
+        """
+        self.x_win.configure(width=width, height=height)
+        self.x_win.change_attributes(
+            win_gravity=X.NorthWestGravity, bit_gravity=X.StaticGravity)
+        # A call to get the window geometry commits the changes.
+        self.geometry
