@@ -50,38 +50,36 @@ from __future__ import absolute_import
 
 import logging
 import os
-import psutil
-import signal
-import subprocess
+import six
 
 from testscenarios import TestWithScenarios
 from testtools import TestCase
-from testtools.content import text_content
 from testtools.matchers import Equals
-from time import sleep
 
-from autopilot.process import ProcessManager
+from autopilot.application import (
+    ClickApplicationLauncher,
+    get_application_launcher_wrapper,
+    NormalApplicationLauncher,
+    UpstartApplicationLauncher,
+)
+from autopilot.display import Display
+from autopilot.globals import get_debug_profile_fixture
 from autopilot.input import Keyboard, Mouse
 from autopilot.introspection import (
-    get_application_launcher,
-    get_application_launcher_from_string_hint,
-    get_autopilot_proxy_object_for_process,
     get_proxy_object_for_existing_process,
-    launch_application,
 )
-from autopilot.introspection.utilities import _get_click_app_id
-from autopilot.display import Display
-from autopilot.utilities import on_test_started
 from autopilot.keybindings import KeybindingsHelper
 from autopilot.matchers import Eventually
+from autopilot.process import ProcessManager
+from autopilot.utilities import deprecated, on_test_started
+from autopilot._timeout import Timeout
 try:
     from autopilot import tracepoint as tp
     HAVE_TRACEPOINT = True
 except ImportError:
     HAVE_TRACEPOINT = False
 
-
-logger = logging.getLogger(__name__)
+_logger = logging.getLogger(__name__)
 
 
 try:
@@ -114,7 +112,7 @@ def _lttng_trace_test_started(test_id):
     if HAVE_TRACEPOINT:
         tp.emit_test_started(test_id)
     else:
-        logger.warning(
+        _logger.warning(
             "No tracing available - install the python-autopilot-trace "
             "package!")
 
@@ -125,6 +123,7 @@ def _lttng_trace_test_ended(test_id):
 
 
 class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
+
     """Wrapper around testtools.TestCase that adds significant functionality.
 
     This class should be the base class for all autopilot test case classes.
@@ -137,6 +136,7 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
     def setUp(self):
         super(AutopilotTestCase, self).setUp()
         on_test_started(self)
+        self.useFixture(get_debug_profile_fixture()(self.addDetail))
 
         _lttng_trace_test_started(self.id())
         self.addCleanup(_lttng_trace_test_ended, self.id())
@@ -146,12 +146,14 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
         self._kb = None
         self._display = None
 
+        # Work around for bug lp:1297592.
+        _ensure_uinput_device_created()
+
         try:
-            self._app_snapshot = \
-                self.process_manager.get_running_applications()
+            self._app_snapshot = _get_process_snapshot()
             self.addCleanup(self._compare_system_with_app_snapshot)
         except RuntimeError:
-            logger.warning(
+            _logger.warning(
                 "Process manager backend unavailable, application snapshot "
                 "support disabled.")
 
@@ -220,11 +222,6 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
             app_proxy = self.launch_test_application(
                 'my_gtk_app.py', app_type='gtk')
 
-        .. seealso::
-
-            Method :py:meth:`AutopilotTestCase.pick_app_launcher`
-                Specify application introspection type globally.
-
         :param application: The application to launch. The application can be
             specified as:
 
@@ -249,42 +246,25 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
         :keyword emulator_base: If set, specifies the base class to be used for
             all emulators for this loaded application.
 
-        :raises: **ValueError** if unknown keyword arguments are passed.
+        :raises ValueError: if unknown keyword arguments are passed.
         :return: A proxy object that represents the application. Introspection
          data is retrievable via this object.
 
         """
-        app_path = subprocess.check_output(['which', application]).strip()
-        # Get a launcher, tests can override this if they need:
-        launcher_hint = kwargs.pop('app_type', '')
-        launcher = None
-        if launcher_hint != '':
-            launcher = get_application_launcher_from_string_hint(launcher_hint)
-        if launcher is None:
-            try:
-                launcher = self.pick_app_launcher(app_path)
-            except RuntimeError:
-                pass
-        if launcher is None:
-            raise RuntimeError(
-                "Autopilot could not determine the correct introspection type "
-                "to use. You can specify one by overriding the "
-                "AutopilotTestCase.pick_app_launcher method.")
-        emulator_base = kwargs.pop('emulator_base', None)
-        dbus_bus = kwargs.pop('dbus_bus', 'session')
-
-        if dbus_bus != 'session':
-            self.patch_environment("DBUS_SESSION_BUS_ADDRESS", dbus_bus)
-
-        process = launch_application(launcher, app_path, *arguments, **kwargs)
-        self.addCleanup(self._kill_process_and_attach_logs, process)
-        return get_autopilot_proxy_object_for_process(
-            process,
-            emulator_base,
-            dbus_bus
+        _logger.info(
+            "Attempting to launch application '%s' with arguments '%s' as a "
+            "normal process",
+            application,
+            ' '.join(arguments)
+        )
+        launcher = self.useFixture(
+            NormalApplicationLauncher(self.addDetail, **kwargs)
         )
 
-    def launch_click_package(self, package_id, app_name=None, **kwargs):
+        return self._launch_test_application(launcher, application, *arguments)
+
+    def launch_click_package(self, package_id, app_name=None, app_uris=[],
+                             **kwargs):
         """Launch a click package application with introspection enabled.
 
         This method takes care of launching a click package with introspection
@@ -303,6 +283,8 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
         :param app_name: Currently, only one application can be packaged in a
             click package, and this parameter can be left at None. If
             specified, it should be the application name you wish to launch.
+        :param app_uris: Parameters used to launch the click package. This
+            parameter will be left empty if not used.
 
         :keyword emulator_base: If set, specifies the base class to be used for
             all emulators for this loaded application.
@@ -312,67 +294,81 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
         :raises RuntimeError: If the specified app_name cannot be found within
             the specified click package.
 
+        :returns: proxy object for the launched package application
+
         """
-        app_id = _get_click_app_id(package_id, app_name)
-        # sadly, we cannot re-use the existing launch_test_application
-        # since upstart is a little odd.
-        # set the qt testability env:
-        subprocess.call([
-            "/sbin/initctl",
-            "set-env",
-            "QT_LOAD_TESTABILITY=1",
-        ])
-        # launch the application:
-        subprocess.check_output([
-            "/sbin/start",
-            "application",
-            "APP_ID={}".format(app_id),
-        ])
-        # perhaps we should do this with a regular expression instead?
-        for i in range(10):
-            try:
-                list_output = subprocess.check_output([
-                    "/sbin/initctl",
-                    "status",
-                    "application-click",
-                    "APP_ID={}".format(app_id)
-                ])
-            except subprocess.CalledProcessError:
-                # application not started yet.
-                pass
-            else:
-                for line in list_output.split('\n'):
-                    if app_id in line and "start/running" in line:
-                        target_pid = int(line.split()[-1])
+        if isinstance(app_uris, (six.text_type, six.binary_type)):
+            app_uris = [app_uris]
+        _logger.info(
+            "Attempting to launch click application '%s' from click package "
+            " '%s' and URIs '%s'",
+            app_name if app_name is not None else "(default)",
+            package_id,
+            ','.join(app_uris)
+        )
+        launcher = self.useFixture(
+            ClickApplicationLauncher(self.addDetail, **kwargs)
+        )
+        return self._launch_test_application(launcher, package_id, app_name,
+                                             app_uris)
 
-                        self.addCleanup(self._kill_pid, target_pid)
-                        logger.info(
-                            "Click package %s has been launched with PID %d",
-                            app_id,
-                            target_pid
-                        )
+    def launch_upstart_application(self, application_name, uris=[], **kwargs):
+        """Launch an application with upstart.
 
-                        emulator_base = kwargs.pop('emulator_base', None)
-                        proxy = get_proxy_object_for_existing_process(
-                            pid=target_pid,
-                            emulator_base=emulator_base
-                        )
-                        # reset the upstart env, and hope no one else launched,
-                        # or they'll have introspection enabled as well,
-                        # although this isn't the worth thing in the world.
-                        subprocess.call([
-                            "/sbin/initctl",
-                            "unset-env",
-                            "QT_LOAD_TESTABILITY",
-                        ])
-                        return proxy
-            # give the app time to launch - maybe this is not needed?:
-            sleep(1)
-        else:
-            raise RuntimeError(
-                "Could not find autopilot interface for click package"
-                " '{}' after 10 seconds.".format(app_id)
-            )
+        This method launched an application via the ``upstart-app-launch``
+        library, on platforms that support it.
+
+        Usage is similar to the
+        :py:meth:`AutopilotTestCase.launch_test_application`::
+
+            app_proxy = self.launch_upstart_application("gallery-app")
+
+        :param application_name: The name of the application to launch.
+        :keyword emulator_base: If set, specifies the base class to be used for
+            all emulators for this loaded application.
+
+        :raises RuntimeError: If the specified application cannot be launched.
+        :raises ValueError: If unknown keyword arguments are specified.
+        """
+        if isinstance(uris, (six.text_type, six.binary_type)):
+            uris = [uris]
+        _logger.info(
+            "Attempting to launch application '%s' with URIs '%s' via "
+            "upstart-app-launch",
+            application_name,
+            ','.join(uris)
+        )
+        launcher = self.useFixture(
+            UpstartApplicationLauncher(self.addDetail, **kwargs)
+        )
+        return self._launch_test_application(launcher, application_name, uris)
+
+    # Wrapper function tying the newer ApplicationLauncher behaviour with the
+    # previous (to be depreciated) behaviour
+    def _launch_test_application(self, launcher_instance, application, *args):
+
+        dbus_bus = launcher_instance.dbus_bus
+        if dbus_bus != 'session':
+            self.patch_environment("DBUS_SESSION_BUS_ADDRESS", dbus_bus)
+
+        pid = launcher_instance.launch(application, *args)
+        process = getattr(launcher_instance, 'process', None)
+        application_name = getattr(
+            launcher_instance,
+            'dbus_application_name',
+            None
+        )
+
+        proxy_obj = get_proxy_object_for_existing_process(
+            pid=pid,
+            process=process,
+            dbus_bus=dbus_bus,
+            emulator_base=launcher_instance.emulator_base,
+            application_name=application_name,
+        )
+        proxy_obj.set_process(process)
+
+        return proxy_obj
 
     def _compare_system_with_app_snapshot(self):
         """Compare the currently running application with the last snapshot.
@@ -381,22 +377,13 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
         applications currently running that were not running when the snapshot
         was taken.
         """
-        if self._app_snapshot is None:
-            raise RuntimeError("No snapshot to match against.")
-
-        new_apps = []
-        for i in range(10):
-            current_apps = self.process_manager.get_running_applications()
-            new_apps = filter(
-                lambda i: i not in self._app_snapshot, current_apps)
-            if not new_apps:
-                self._app_snapshot = None
-                return
-            sleep(1)
-        self._app_snapshot = None
-        raise AssertionError(
-            "The following apps were started during the test and not closed: "
-            "%r", new_apps)
+        try:
+            _compare_system_with_process_snapshot(
+                _get_process_snapshot,
+                self._app_snapshot
+            )
+        finally:
+            self._app_snapshot = None
 
     def patch_environment(self, key, value):
         """Patch the process environment, setting *key* with value *value*.
@@ -424,7 +411,7 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
         """
         if key in os.environ:
             def _undo_patch(key, old_value):
-                logger.info(
+                _logger.info(
                     "Resetting environment variable '%s' to '%s'",
                     key,
                     old_value
@@ -435,14 +422,24 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
         else:
             def _remove_patch(key):
                 try:
+                    _logger.info(
+                        "Deleting previously-created environment "
+                        "variable '%s'",
+                        key
+                    )
                     del os.environ[key]
                 except KeyError:
-                    logger.warning(
+                    _logger.warning(
                         "Attempted to delete environment key '%s' that doesn't"
                         "exist in the environment",
                         key
                     )
             self.addCleanup(_remove_patch, key)
+        _logger.info(
+            "Setting environment variable '%s' to '%s'",
+            key,
+            value
+        )
         os.environ[key] = value
 
     def assertVisibleWindowStack(self, stack_start):
@@ -453,7 +450,7 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
 
         :param stack_start: An iterable of
          :class:`~autopilot.process.Window` instances.
-        :raises: **AssertionError** if the top of the window stack does not
+        :raises AssertionError: if the top of the window stack does not
          match the contents of the stack_start parameter.
 
         """
@@ -483,16 +480,16 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
         :param obj: The object to test.
         :param kwargs: One or more keyword arguments to match against the
          attributes of the *obj* parameter.
-        :raises: **ValueError** if no keyword arguments were given.
-        :raises: **ValueError** if a named attribute is a callable object.
-        :raises: **AssertionError** if any of the attribute/value pairs in
+        :raises ValueError: if no keyword arguments were given.
+        :raises ValueError: if a named attribute is a callable object.
+        :raises AssertionError: if any of the attribute/value pairs in
          kwargs do not match the attributes on the object passed in.
 
         """
         if not kwargs:
             raise ValueError("At least one keyword argument must be present.")
 
-        for prop_name, desired_value in kwargs.iteritems():
+        for prop_name, desired_value in kwargs.items():
             none_val = object()
             attr = getattr(obj, prop_name, none_val)
             if attr == none_val:
@@ -509,6 +506,9 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
 
     assertProperties = assertProperty
 
+    @deprecated(
+        "the 'app_type' argument to the launch_test_application method"
+    )
     def pick_app_launcher(self, app_path):
         """Given an application path, return an object suitable for launching
         the application.
@@ -519,63 +519,54 @@ class AutopilotTestCase(TestWithScenarios, TestCase, KeybindingsHelper):
         own implemetnation.
 
         The default implementation calls
-        :py:func:`autopilot.introspection.get_application_launcher`
+        :py:func:`autopilot.application.get_application_launcher_wrapper`
 
         """
-        # default implementation is in autopilot.introspection:
-        return get_application_launcher(app_path)
-
-    def _kill_pid(self, pid):
-        """Kill the process with the specified pid."""
-        logger.info("waiting for process to exit.")
-        try:
-            logger.info("Killing process %d", pid)
-            os.killpg(pid, signal.SIGTERM)
-        except OSError:
-            logger.info("Appears process has already exited.")
-        for i in range(10):
-            if not _is_process_running(pid):
-                break
-            if i == 9:
-                logger.info(
-                    "Killing process group, since it hasn't exited after "
-                    "10 seconds."
-                )
-                os.killpg(pid, signal.SIGKILL)
-            sleep(1)
-
-    def _kill_process(self, process):
-        """Kill the process, and return the stdout, stderr and return code."""
-        stdout = ""
-        stderr = ""
-        return_code = -1
-        logger.info("waiting for process to exit.")
-        try:
-            logger.info("Killing process %d", process.pid)
-            os.killpg(process.pid, signal.SIGTERM)
-        except OSError:
-            logger.info("Appears process has already exited.")
-        for i in range(10):
-            tmp_out, tmp_err = process.communicate()
-            stdout += tmp_out
-            stderr += tmp_err
-            if not _is_process_running(process.pid):
-                break
-            if i == 9:
-                logger.info(
-                    "Killing process group, since it hasn't exited after "
-                    "10 seconds."
-                )
-                os.killpg(process.pid, signal.SIGKILL)
-            sleep(1)
-        return stdout, stderr, process.returncode
-
-    def _kill_process_and_attach_logs(self, process):
-        stdout, stderr, return_code = self._kill_process(process)
-        self.addDetail('process-return-code', text_content(str(return_code)))
-        self.addDetail('process-stdout', text_content(stdout))
-        self.addDetail('process-stderr', text_content(stderr))
+        # default implementation is in autopilot.application:
+        return get_application_launcher_wrapper(app_path)
 
 
-def _is_process_running(pid):
-    return pid in psutil.get_pid_list()
+def _get_process_snapshot():
+    """Return a snapshot of running processes on the system.
+
+    :returns: a list of running processes.
+    :raises RuntimeError: if the process manager is unsavailble on this
+        platform.
+
+    """
+    return ProcessManager.create().get_running_applications()
+
+
+def _compare_system_with_process_snapshot(snapshot_fn, old_snapshot):
+    """Compare an existing process snapshot with current running processes.
+
+    :param snapshot_fn: A callable that returns the current running process
+        list.
+    :param old_snapshot: A list of processes to compare against.
+    :raises AssertionError: If, after 10 seconds, there are still running
+        processes that were not present in ``old_snapshot``.
+
+    """
+    new_apps = []
+    for _ in Timeout.default():
+        current_apps = snapshot_fn()
+        new_apps = [app for app in current_apps if app not in old_snapshot]
+        if not new_apps:
+            return
+    raise AssertionError(
+        "The following apps were started during the test and not closed: "
+        "%r" % new_apps)
+
+
+def _ensure_uinput_device_created():
+    # This exists for a work around for bug lp:1297592. Need to create
+    # an input device before an application launch.
+    try:
+        from autopilot.input._uinput import Touch, _UInputTouchDevice
+        if _UInputTouchDevice._device is None:
+            Touch.create()
+    except Exception as e:
+        _logger.warning(
+            "Failed to create Touch device for bug lp:1297595 workaround: "
+            "%s" % str(e)
+        )
