@@ -20,13 +20,15 @@
 
 from __future__ import absolute_import
 
-from codecs import open
 import os
 import os.path
+import re
 from tempfile import mktemp
+from testtools import skipIf
 from testtools.matchers import Contains, Equals, MatchesRegex, Not
 from textwrap import dedent
 
+from autopilot import platform
 from autopilot.testcase import AutopilotTestCase
 from autopilot.tests.functional import AutopilotRunTestBase, remove_if_exists
 
@@ -48,23 +50,44 @@ class AutopilotFunctionalTestsBase(AutopilotRunTestBase):
         return self.run_autopilot(args_list)
 
     def assertTestsInOutput(self, tests, output, total_title='tests'):
-        """Asserts that 'tests' are all present in 'output'."""
+        """Asserts that 'tests' are all present in 'output'.
+
+        This assertion is intelligent enough to know that tests are not always
+        printed in alphabetical order.
+
+        'tests' can either be a list of test ids, or a list of tuples
+        containing (scenario_count, test_id), in the case of scenarios.
+
+        """
 
         if type(tests) is not list:
             raise TypeError("tests must be a list, not %r" % type(tests))
-        if not isinstance(output, basestring):
+        if not isinstance(output, str):
             raise TypeError("output must be a string, not %r" % type(output))
 
-        test_names = ''.join(['    %s\n' % t for t in sorted(tests)])
-        expected = '''\
-Loading tests from: %s
+        expected_heading = 'Loading tests from: %s\n\n' % self.base_path
+        expected_tests = []
+        expected_total = 0
+        for test in tests:
+            if type(test) == tuple:
+                expected_tests.append(' *%d %s' % test)
+                expected_total += test[0]
+            else:
+                expected_tests.append('    %s' % test)
+                expected_total += 1
+        expected_footer = ' %d total %s.' % (expected_total, total_title)
 
-%s
+        parts = output.split('\n')
+        observed_heading = '\n'.join(parts[:2]) + '\n'
+        observed_test_list = parts[2:-4]
+        observed_footer = parts[-2]
 
- %d total %s.
-''' % (self.base_path, test_names, len(tests), total_title)
-
-        self.assertThat(output, Equals(expected))
+        self.assertThat(expected_heading, Equals(observed_heading))
+        self.assertThat(
+            sorted(expected_tests),
+            Equals(sorted(observed_test_list))
+        )
+        self.assertThat(expected_footer, Equals(observed_footer))
 
     def test_can_list_empty_test_dir(self):
         """Autopilot list must report 0 tests found with an empty test
@@ -119,10 +142,15 @@ Loading tests from: %s
             """)
         )
         code, output, error = self.run_autopilot_list()
-        expected_output = '''ImportError: No module named asdjkhdfjgsdhfjhsd'''
         self.assertThat(code, Equals(0))
         self.assertThat(error, Equals(''))
-        self.assertThat(output, Contains(expected_output))
+        self.assertThat(
+            output,
+            MatchesRegex(
+                ".*ImportError: No module named [']?asdjkhdfjgsdhfjhsd[']?.*",
+                re.DOTALL
+            )
+        )
 
     def test_list_tests_with_syntax_error(self):
         self.create_test_file(
@@ -144,6 +172,14 @@ Loading tests from: %s
         self.assertThat(error, Equals(''))
         self.assertThat(output, Contains(expected_error))
 
+    def test_list_nonexistent_test_returns_nonzero(self):
+        code, output, error = self.run_autopilot_list(list_spec='1234')
+        expected_msg = "could not import package 1234: No module"
+        expected_result = "0 total tests"
+        self.assertThat(code, Equals(1))
+        self.assertThat(output, Contains(expected_msg))
+        self.assertThat(output, Contains(expected_result))
+
     def test_can_list_scenariod_tests(self):
         """Autopilot must show scenario counts next to tests that have
         scenarios."""
@@ -207,20 +243,16 @@ Loading tests from: %s
             """)
         )
 
-        expected_output = '''\
-Loading tests from: %s
-
- *2 tests.test_simple.SimpleTest.test_simple
- *2 tests.test_simple.SimpleTest.test_simple_two
-
-
- 4 total tests.
-''' % self.base_path
-
         code, output, error = self.run_autopilot_list()
         self.assertThat(code, Equals(0))
         self.assertThat(error, Equals(''))
-        self.assertThat(output, Equals(expected_output))
+        self.assertTestsInOutput(
+            [
+                (2, 'tests.test_simple.SimpleTest.test_simple'),
+                (2, 'tests.test_simple.SimpleTest.test_simple_two'),
+            ],
+            output
+        )
 
     def test_can_list_invalid_scenarios(self):
         """Autopilot must ignore scenarios that are not lists."""
@@ -245,6 +277,20 @@ Loading tests from: %s
         self.assertTestsInOutput(
             ['tests.test_simple.SimpleTest.test_simple'], output)
 
+    def test_local_module_loaded_and_not_system_module(self):
+        module_path1 = self.create_empty_test_module()
+        module_path2 = self.create_empty_test_module()
+
+        self.base_path = module_path2
+
+        retcode, stdout, stderr = self.run_autopilot(
+            ["run", "tests"],
+            pythonpath=module_path1,
+            use_script=True
+        )
+
+        self.assertThat(stdout, Contains(module_path2))
+
     def test_can_list_just_suites(self):
         """Must only list available suites, not the contained tests."""
         self.create_test_file(
@@ -276,6 +322,7 @@ Loading tests from: %s
              'tests.test_simple_suites.AnotherSimpleTest'],
             output, total_title='suites')
 
+    @skipIf(platform.model() != "Desktop", "Only suitable on Desktop (VidRec)")
     def test_record_flag_works(self):
         """Must be able to record videos when the -r flag is present."""
 
@@ -313,6 +360,7 @@ Loading tests from: %s
         if should_delete:
             self.addCleanup(remove_if_exists, "/tmp/autopilot")
 
+    @skipIf(platform.model() != "Desktop", "Only suitable on Desktop (VidRec)")
     def test_record_dir_option_and_record_works(self):
         """Must be able to specify record directory flag and record."""
 
@@ -356,6 +404,7 @@ Loading tests from: %s
             os.path.exists(
                 '%s/tests.test_simple.SimpleTest.test_simple.ogv' % (ap_dir)))
 
+    @skipIf(platform.model() != "Desktop", "Only suitable on Desktop (VidRec)")
     def test_record_dir_option_works(self):
         """Must be able to specify record directory flag."""
 
@@ -389,6 +438,7 @@ Loading tests from: %s
                 '%s/tests.test_simple.SimpleTest.test_simple.ogv' %
                 (video_dir)))
 
+    @skipIf(platform.model() != "Desktop", "Only suitable on Desktop (VidRec)")
     def test_no_videos_saved_when_record_option_is_not_present(self):
         """Videos must not be saved if the '-r' option is not specified."""
         self.create_test_file(
@@ -414,6 +464,7 @@ Loading tests from: %s
         self.assertFalse(os.path.exists(
             '/tmp/autopilot/tests.test_simple.SimpleTest.test_simple.ogv'))
 
+    @skipIf(platform.model() != "Desktop", "Only suitable on Desktop (VidRec)")
     def test_no_videos_saved_for_skipped_test(self):
         """Videos must not be saved if the test has been skipped (not
         failed).
@@ -442,6 +493,7 @@ Loading tests from: %s
         self.assertThat(code, Equals(0))
         self.assertThat(os.path.exists(video_file_path), Equals(False))
 
+    @skipIf(platform.model() != "Desktop", "Only suitable on Desktop (VidRec)")
     def test_no_video_for_nested_testcase_when_parent_and_child_fail(self):
         """Test recording must not create a new recording for nested testcases
         where both the parent and the child testcase fail.
@@ -500,11 +552,15 @@ Loading tests from: %s
 
         code, output, error = self.run_autopilot(["run", "tests"])
 
-        expected_error = 'ImportError: No module named asdjkhdfjgsdhfjhsd'
-
         self.assertThat(code, Equals(1))
         self.assertThat(error, Equals(''))
-        self.assertThat(output, Contains(expected_error))
+        self.assertThat(
+            output,
+            MatchesRegex(
+                ".*ImportError: No module named [']?asdjkhdfjgsdhfjhsd[']?.*",
+                re.DOTALL
+            )
+        )
         self.assertThat(output, Contains("FAILED (failures=1)"))
 
     def test_runs_with_syntax_errors_fail(self):
@@ -538,54 +594,17 @@ SyntaxError: invalid syntax
         self.assertThat(output, Contains(expected_error))
         self.assertThat(output, Contains("FAILED (failures=1)"))
 
-    def test_can_error_with_unicode_data(self):
-        """Tests that assert with unicode errors must get saved to a log
-        file."""
+    def test_can_create_subunit_result_file(self):
         self.create_test_file(
             "test_simple.py", dedent(u"""\
-            # encoding: utf-8
 
-            # from autopilot.testcase import AutopilotTestCase
-            from testtools import TestCase
+            from autopilot.testcase import AutopilotTestCase
 
-            class SimpleTest(TestCase):
+
+            class SimpleTest(AutopilotTestCase):
 
                 def test_simple(self):
-                    self.fail(
-                        u'\xa1pl\u0279oM \u01ddpo\u0254\u0131u\u2229 oll'
-                        u'\u01ddH')
-
-            """)
-        )
-        output_file_path = mktemp()
-        self.addCleanup(remove_if_exists, output_file_path)
-
-        code, output, error = self.run_autopilot(
-            ["run", "-o", output_file_path, "tests"])
-
-        self.assertThat(code, Equals(1))
-        self.assertTrue(os.path.exists(output_file_path))
-        log_contents = unicode(open(output_file_path, encoding='utf-8').read())
-        self.assertThat(
-            log_contents,
-            Contains(u'\xa1pl\u0279oM \u01ddpo\u0254\u0131u\u2229 oll\u01ddH'))
-
-    def test_can_write_xml_error_with_unicode_data(self):
-        """Tests that assert with unicode errors must get saved to XML log
-        file."""
-        self.create_test_file(
-            "test_simple.py", dedent(u"""\
-            # encoding: utf-8
-
-            # from autopilot.testcase import AutopilotTestCase
-            from testtools import TestCase
-
-            class SimpleTest(TestCase):
-
-                def test_simple(self):
-                    error = (u'\xa1pl\u0279oM \u01ddpo\u0254\u0131u'
-                        u'\u2229 oll\u01ddH')
-                    self.fail(error)
+                    pass
 
             """)
         )
@@ -595,15 +614,11 @@ SyntaxError: invalid syntax
         code, output, error = self.run_autopilot([
             "run",
             "-o", output_file_path,
-            "-f", "xml",
+            "-f", "subunit",
             "tests"])
 
-        self.assertThat(code, Equals(1))
+        self.assertThat(code, Equals(0))
         self.assertTrue(os.path.exists(output_file_path))
-        log_contents = unicode(open(output_file_path, encoding='utf-8').read())
-        self.assertThat(
-            log_contents,
-            Contains(u'\xa1pl\u0279oM \u01ddpo\u0254\u0131u\u2229 oll\u01ddH'))
 
     def test_launch_needs_arguments(self):
         """Autopilot launch must complain if not given an application to
@@ -632,7 +647,7 @@ SyntaxError: invalid syntax
 
         self.assertThat(rc, Equals(1))
         self.assertThat(
-            stdout, Contains("Error: cannot find application 'DoEsNotExist'"))
+            stdout, Contains("Error: Cannot find application 'DoEsNotExist'"))
 
     def test_complains_on_non_dynamic_binary(self):
         """Must give a nice error message when passing in a non-dynamic
@@ -825,7 +840,7 @@ class AutopilotVerboseFunctionalTests(AutopilotFunctionalTestsBase):
             class SimpleTest(AutopilotTestCase):
 
                 def test_simple(self):
-                    self.assertTrue()
+                    raise RuntimeError("Intentionally fail test.")
             """)
         )
 
@@ -837,8 +852,9 @@ class AutopilotVerboseFunctionalTests(AutopilotFunctionalTestsBase):
             error, Contains("ERROR: tests.test_simple.SimpleTest.test_simple"))
         self.assertThat(error, Contains("traceback:"))
         self.assertThat(
-            error, Contains("TypeError: assertTrue() takes at least 2 "
-                            "arguments (1 given)"))
+            error,
+            Contains("RuntimeError: Intentionally fail test.")
+        )
 
     def test_verbose_flag_shows_failure(self):
         """Verbose log must indicate a test failure with a traceback (xml
@@ -965,3 +981,27 @@ class AutopilotVerboseFunctionalTests(AutopilotFunctionalTestsBase):
                                                   "-v", "tests"])
         self.assertThat(
             error, Contains(get_version_string()))
+
+    def test_failfast(self):
+        """Run stops after first error encountered."""
+        self.create_test_file(
+            'test_failfast.py', dedent("""\
+
+            from autopilot.testcase import AutopilotTestCase
+
+
+            class SimpleTest(AutopilotTestCase):
+
+                def test_one(self):
+                    raise Exception
+
+                def test_two(self):
+                    raise Exception
+            """)
+        )
+        code, output, error = self.run_autopilot(["run",
+                                                  "--failfast",
+                                                  "tests"])
+        self.assertThat(code, Equals(1))
+        self.assertIn("Ran 1 test", output)
+        self.assertIn("FAILED (failures=1)", output)
