@@ -21,11 +21,13 @@
 from __future__ import absolute_import
 
 import dbus
+import logging
 from PyQt4 import QtGui, QtCore
+import six
 
 from autopilot.introspection import (
     _get_dbus_address_object,
-    _make_proxy_object
+    _make_proxy_object_async
 )
 from autopilot.introspection.constants import AP_INTROSPECTION_IFACE
 from autopilot.introspection.dbus import StateNotFoundError
@@ -33,6 +35,8 @@ from autopilot.introspection.qt import QtObjectProxyMixin
 from autopilot.vis.objectproperties import TreeNodeDetailWidget
 from autopilot.vis.resources import get_qt_icon
 
+_logger = logging.getLogger(__name__)
+
 
 class MainWindow(QtGui.QMainWindow):
     def __init__(self, dbus_bus):
@@ -44,8 +48,12 @@ class MainWindow(QtGui.QMainWindow):
 
     def readSettings(self):
         settings = QtCore.QSettings()
-        self.restoreGeometry(settings.value("geometry").toByteArray())
-        self.restoreState(settings.value("windowState").toByteArray())
+        if six.PY3:
+            self.restoreGeometry(settings.value("geometry").data())
+            self.restoreState(settings.value("windowState").data())
+        else:
+            self.restoreGeometry(settings.value("geometry").toByteArray())
+            self.restoreState(settings.value("windowState").toByteArray())
 
     def closeEvent(self, event):
         settings = QtCore.QSettings()
@@ -56,6 +64,7 @@ class MainWindow(QtGui.QMainWindow):
         self.statusBar().showMessage('Waiting for first valid dbus connection')
 
         self.splitter = QtGui.QSplitter(self)
+        self.splitter.setChildrenCollapsible(False)
         self.tree_view = QtGui.QTreeView(self.splitter)
         self.detail_widget = TreeNodeDetailWidget(self.splitter)
 
@@ -78,29 +87,41 @@ class MainWindow(QtGui.QMainWindow):
             try:
                 dbus_address_instance = _get_dbus_address_object(
                     str(conn), str(obj), self._dbus_bus)
-                proxy_object = _make_proxy_object(dbus_address_instance, None)
-                cls_name = proxy_object.__class__.__name__
-                if not cls_name in self.selectable_interfaces:
-                    self.selectable_interfaces[cls_name] = proxy_object
-                    self.update_selectable_interfaces()
-            except (dbus.DBusException, RuntimeError):
-                pass
-            self.statusBar().clearMessage()
+                _make_proxy_object_async(
+                    dbus_address_instance,
+                    None,
+                    self.on_proxy_object_built,
+                    self.on_dbus_error
+                )
+            except (dbus.DBusException, RuntimeError) as e:
+                _logger.warning("Invalid introspection interface: %s" % str(e))
+
+            if self.connection_list.count() == 0:
+                self.statusBar().showMessage('No valid connections exist.')
+
+    def on_proxy_object_built(self, proxy_object):
+        cls_name = proxy_object.__class__.__name__
+        if not cls_name in self.selectable_interfaces:
+            self.selectable_interfaces[cls_name] = proxy_object
+            self.update_selectable_interfaces()
+        self.statusBar().clearMessage()
+
+    def on_dbus_error(*args):
+        print(args)
 
     def update_selectable_interfaces(self):
         selected_text = self.connection_list.currentText()
         self.connection_list.clear()
-        self.connection_list.addItem("Please select a connection",
-                                     QtCore.QVariant(None))
-        for name, proxy_obj in self.selectable_interfaces.iteritems():
+        self.connection_list.addItem("Please select a connection", None)
+        for name, proxy_obj in self.selectable_interfaces.items():
             if isinstance(proxy_obj, QtObjectProxyMixin):
                 self.connection_list.addItem(
                     get_qt_icon(),
                     name,
-                    QtCore.QVariant(proxy_obj)
+                    proxy_obj
                 )
             else:
-                self.connection_list.addItem(name, QtCore.QVariant(proxy_obj))
+                self.connection_list.addItem(name, proxy_obj)
 
         prev_selected = self.connection_list.findText(selected_text,
                                                       QtCore.Qt.MatchExactly)
@@ -109,7 +130,9 @@ class MainWindow(QtGui.QMainWindow):
         self.connection_list.setCurrentIndex(prev_selected)
 
     def conn_list_activated(self, index):
-        dbus_details = self.connection_list.itemData(index).toPyObject()
+        dbus_details = self.connection_list.itemData(index)
+        if not six.PY3:
+            dbus_details = dbus_details.toPyObject()
         if dbus_details:
             self.tree_model = VisTreeModel(dbus_details)
             self.tree_view.setModel(self.tree_model)
@@ -211,16 +234,16 @@ class VisTreeModel(QtCore.QAbstractItemModel):
 
     def data(self, index, role):
         if not index.isValid():
-            return QtCore.QVariant()
+            return None
         if role == QtCore.Qt.DisplayRole:
-            return QtCore.QVariant(index.internalPointer().name)
+            return index.internalPointer().name
 
     def headerData(self, column, orientation, role):
         if (orientation == QtCore.Qt.Horizontal and
                 role == QtCore.Qt.DisplayRole):
             try:
-                return QtCore.QVariant("Tree Node")
+                return "Tree Node"
             except IndexError:
                 pass
 
-        return QtCore.QVariant()
+        return None
